//
// 
//

#include "DataContainerSecond.h"



AbstractionExtension::AbstractionExtension(AbstractionAction action, std::vector<int> events, int id) {
//    if (action == DO_NOTHING){
//        throw std::runtime_error("Wrong Constructor called");
//    }
    this->action = action;
    this->events = std::move(events);
    this->id = id;

}

AbstractionExtension::AbstractionExtension() {
    this->action = DO_NOTHING;
}

AbstractionExtension::AbstractionExtension(AbstractionAction action, std::vector<int> events, int id,
                                           UniqueValueCounter<int> frequUpdates,
                                           std::map<std::pair<int, int>, std::vector<int>> absUpdates) {
    if (action != EXTEND){
        throw std::runtime_error("Wrong Constructor called");
    }
    this->action = action;
    this->events = std::move(events);
    this->id = id;
    this->frequUpdates = std::move(frequUpdates);
    this->absUpdates = std::move(absUpdates);


}

void AbstractionExtension::print() {
    std::cout << "Id: " << id << " Action: " << action << " Events: ";
    for (auto e:events){
        std::cout << e << ", ";
    }
    std::cout << std::endl;
}

ExtendCandidateReturn::ExtendCandidateReturn(double estimatedGain, Pattern *source_pattern,
                                             Pattern* best_extension,
                                             std::vector<AbstractionExtension> abstractionExtensions) : best_extension(best_extension), pExt(PatternExtension(*source_pattern)) {
    this->estimatedGain = estimatedGain;
    this->source_pattern = source_pattern;
    this->abstractionExtensions= std::move(abstractionExtensions);
    this->contains_pExt = false;

}


ExtendCandidateReturn::ExtendCandidateReturn(double estimatedGain, Pattern *source_pattern, Pattern *best_extension,
                                             std::vector<AbstractionExtension> abstractionExtensions,
                                             PatternExtension pExt) : pExt(pExt), best_extension(best_extension) {
    this->estimatedGain = estimatedGain;
    this->source_pattern = source_pattern;
    this->abstractionExtensions= std::move(abstractionExtensions);
    this->contains_pExt = true;
}

bool ExtendCandidateReturn::operator<(const ExtendCandidateReturn &rhs) const {
    return (this->estimatedGain < rhs.estimatedGain);
}


LocationExtension::LocationExtension(int relativLocation) {
    this->relativLocation = relativLocation;
}

PatternExtension::PatternExtension(Pattern& pattern, bool interleave) : pattern(pattern) {
    this->interleave = interleave;
}

void PatternExtension::addExtension(int event, int relativlocation, ExtensionKind kind) {
    if (interleave){
        for (auto& e:extensions){
            if(e.relativLocation == relativlocation){
                if (kind == KIND_SPLICE){
                    throw std::runtime_error("Can't splice on existing location when interleave");
                }else if (kind == KIND_REPLACE){
                    e.event = event;
                    return;
                }
            }
        }
        if (kind == KIND_SPLICE){
            auto pe = LocationExtension(relativlocation);
            pe.event = event;
            extensions.push_back(pe);
        }else if (kind == KIND_REPLACE){
            throw std::runtime_error("Can't replace event that does not exits!");
        }
    }else{
        bool added = false;
        if (kind == KIND_SPLICE){
            for (auto& e:extensions){
                if (relativlocation >0 && e.relativLocation >= relativlocation){
                    e.relativLocation += 1;
                }
                if (relativlocation < 0 && e.relativLocation <= relativlocation){
                    e.relativLocation -= 1;
                }
            }
            auto pe = LocationExtension(relativlocation);
            pe.event = event;
            extensions.push_back(pe);
            added = true;
        }else if (kind == KIND_REPLACE){
            for (auto& e:extensions){
                if (e.relativLocation == relativlocation){
                    e.event = event;
                    added = true;
                }
            }
        }
        if (!added){
            throw std::runtime_error("Event was not added!");
        }
    }
}

LocationExtension& PatternExtension::getLocationExtension(int relativPosition) {
    for (auto& e:extensions){
        if (e.relativLocation == relativPosition){
            return e;
        }
    }
    throw std::runtime_error("Location not found!");
}

int PatternExtension::getPatternSize() {
    return pattern.getEvents().size() + extensions.size();
}

Pattern *PatternExtension::buildPattern() {
    if (interleave){
        auto p = new Pattern(pattern);
        int shift = 0;
        std::sort(extensions.begin(),extensions.end(), less_than_key_LocationExt());
        for (auto e:extensions){
            LinkType linkType;
            if (e.event > GlobalData::db->getHighestEvent()){// is abstaction
                linkType = DEPENDENT_ON;
            } else{
                linkType = NONE;
            }
            p->replace(e.relativLocation+1+shift, e.event,-1, linkType, true );
            shift++;
        }
        return p;
    }else{
        auto p = new Pattern(pattern);
        std::sort(extensions.begin(),extensions.end(), less_than_key_LocationExt());
        int position = 0;
        bool extend_left_size = true;
        for (auto e:extensions){
            if (e.relativLocation > 0 && extend_left_size){
                position += pattern.getEvents().size();
                extend_left_size = false;
            }
            LinkType linkType;
            if (e.event > GlobalData::db->getHighestEvent()){// is abstaction
                linkType = DEPENDENT_ON;
            } else{
                linkType = NONE;
            }
            p->replace(position,e.event,-1,linkType,true);
            position++;
        }
        return p;
    }
}

void PatternExtension::print() {
    std::sort(extensions.begin(),extensions.end(),less_than_key_LocationExt());
    std::cout << "Pattern: ";
    pattern.print();
    std::cout << "Pattern Extensions: " ;
    for (auto e: extensions){
        std::cout << "(" << e.relativLocation << "," << e.event << ") ";
    }
    std::cout << std::endl;
}

ReturnGainEstAbsExt::ReturnGainEstAbsExt(double delta_in_bits, UniqueValueCounter<int> combined_pattern_frequ_updates,
                                         int added_abstractions) {
    this->delta_in_bits = delta_in_bits;
    this->combined_pattern_frequ_updates = combined_pattern_frequ_updates;
    this->added_abstractions = added_abstractions;
}

PQEntryPatternOverlap::PQEntryPatternOverlap(int p1, int p2, std::vector<int> overlap, int combined_usage) {
    this->p1 = p1;
    this->p2 = p2;
    this->overlap = std::move(overlap);
    this->combined_usage = combined_usage;
    this->overlap_length = this->overlap.size();
}

bool PQEntryPatternOverlap::operator<(const PQEntryPatternOverlap &rhs) const {
    if (overlap_length == rhs.overlap_length){
        return (combined_usage < rhs.combined_usage);
    }else{
        return (overlap_length < rhs.overlap_length);
    }
}

bool PQEntryPatternOverlap::operator>(const PQEntryPatternOverlap &rhs) const {
    if (overlap_length == rhs.overlap_length){
        return (combined_usage > rhs.combined_usage);
    }else{
        return (overlap_length > rhs.overlap_length);
    }
}
