//
// 
//

#ifndef META_DATACONTAINERSECOND_H
#define META_DATACONTAINERSECOND_H



#include "Pattern.h"
#include "GlobalData.h"
#include <utility>


enum AbstractionAction {ADD_NEW, EXTEND, DO_NOTHING};

struct AbstractionExtension{
    /*
     *  Describes how to extend our abstractions
     */
    AbstractionExtension();
    AbstractionExtension(AbstractionAction action, std::vector<int> events, int id);

    AbstractionExtension(AbstractionAction action, std::vector<int> events, int id,
                         UniqueValueCounter<int> frequUpdates,
                         std::map<std::pair<int, int>, std::vector<int>> absUpdates);

    void print();

    AbstractionAction action;
    std::vector<int> events;
    int id;
    UniqueValueCounter<int> frequUpdates;
    std::map<std::pair<int,int>,std::vector<int>> absUpdates;
};


enum ExtensionKind {KIND_SPLICE, KIND_REPLACE};

struct LocationExtension{
    LocationExtension(int relativLocation);
    int relativLocation;
    int event;
};

struct PatternExtension{
    PatternExtension(Pattern& pattern, bool interleave=false);
    void addExtension(int event, int relativlocation, ExtensionKind kind);
    LocationExtension& getLocationExtension(int relativPosition);
    int getPatternSize();
    Pattern* buildPattern();
    void print();

    Pattern pattern;
    std::vector<LocationExtension> extensions;
    bool interleave;
};

struct less_than_key_LocationExt
{
    inline bool operator() (const LocationExtension& lhs, const LocationExtension& rhs)
    {
        return (lhs.relativLocation < rhs.relativLocation);
    }
};

struct ExtendCandidateReturn{
    ExtendCandidateReturn(double estimatedGain, Pattern *source_pattern,
                          Pattern* best_extension,
                          std::vector<AbstractionExtension> abstractionExtensions);
    ExtendCandidateReturn(double estimatedGain, Pattern *source_pattern,
                          Pattern* best_extension,
                          std::vector<AbstractionExtension> abstractionExtensions, PatternExtension pExt);
    double estimatedGain;
    Pattern* best_extension;
    Pattern* source_pattern;
    std::vector<AbstractionExtension> abstractionExtensions;
    bool contains_pExt;
    PatternExtension pExt;

    bool operator<(const ExtendCandidateReturn &rhs) const;
};


struct ReturnGainEstAbsExt{
    ReturnGainEstAbsExt(double delta_in_bits, UniqueValueCounter<int> combined_pattern_frequ_updates, int added_abstractions);
    UniqueValueCounter<int> combined_pattern_frequ_updates;
    double delta_in_bits;
    int added_abstractions;
};

struct PQEntryPatternOverlap{
    PQEntryPatternOverlap(int p1, int p2, std::vector<int> overlap, int combined_usage);

    bool operator<(const PQEntryPatternOverlap& rhs) const;
    bool operator>(const PQEntryPatternOverlap& rhs) const;

    int p1;
    int p2;
    std::vector<int> overlap;
    int combined_usage;
private:
    int overlap_length;
};



#endif //META_DATACONTAINERSECOND_H
