//
// 
//

#include "DataContainers.h"

//#include <utility>


DataIndex::DataIndex(int sequence_id, int event_id) {
    this->sequence_id = sequence_id;
    this->event_id = event_id;
}

bool DataIndex::operator<(const DataIndex &rhs) const {
    if (sequence_id != rhs.sequence_id){
        return (sequence_id < rhs.sequence_id);
    }else{
        return (event_id < rhs.event_id);
    }
}

bool DataIndex::operator==(const DataIndex &rhs) const {
    return (sequence_id == rhs.sequence_id && event_id == rhs.event_id);
}

bool DataIndex::operator>(const DataIndex &rhs) const {
    if (sequence_id != rhs.sequence_id){
        return (sequence_id > rhs.sequence_id);
    }else{
        return (event_id > rhs.event_id);
    }
}

int DataIndex::operator-(const DataIndex &rhs) const {
    if (sequence_id != rhs.sequence_id){
        throw std::runtime_error("Can't compute difference between location of different data sequences.");
        // well you can but there shouldn't be any reason for it.
    } else{
        return event_id - rhs.event_id;
    }
}

bool DataIndex::operator>=(const DataIndex &rhs) const {
    return (*this > rhs || *this == rhs);
}

DataIndex& DataIndex::operator++() {
    this->event_id++;
    return *this;
}


PQEntryFindWin::PQEntryFindWin(DataIndex start_index, int pattern_length)
        :  current_index(start_index), pattern_length(pattern_length) {
    positions.push_back(start_index);
    remaining_symbols = pattern_length - 1; //first symbol is start index
}

bool PQEntryFindWin::operator<(const PQEntryFindWin &rhs) const {
    return (remaining_symbols + gaps)<(rhs.remaining_symbols + gaps);
}

int PQEntryFindWin::getNextEventIndex() {
    return pattern_length - remaining_symbols;
}

void PQEntryFindWin::addPosition(DataIndex nextPosition) {
    positions.push_back(nextPosition);
    remaining_symbols--;
}

std::vector<DataIndex> PQEntryFindWin::getPositions() {
    return positions;
}

PQEntryGapDistributionExtension::PQEntryGapDistributionExtension(int gap_distribution_of,  int relative_to,
                                                                 double distance) {
    this->gap_distribution_of = gap_distribution_of;
    this->relative_to = relative_to;
    this->distance = distance;
}

bool PQEntryGapDistributionExtension::operator<(const PQEntryGapDistributionExtension &rhs) const {
    return (this->distance < rhs.distance);
}

bool PQEntryGapDistributionExtension::operator>(const PQEntryGapDistributionExtension &rhs) const {
    return (this->distance > rhs.distance);
}

void PQEntryGapDistributionExtension::print() {
    std::cout << "Gap Dst Of: " << this->gap_distribution_of << " relative to: " << this->relative_to << " Distance: " << this->distance << std::endl;
}


PQEntryRealCounts::PQEntryRealCounts(int pattern_e_index, int event_id, int counts) {
    this->pattern_e_index = pattern_e_index;
    this->event_id = event_id;
    this->counts = counts;
}

bool PQEntryRealCounts::operator<(const PQEntryRealCounts rhs) const {
    return (this->counts < rhs.counts);
}

bool PQEntryRealCounts::operator>(const PQEntryRealCounts rhs) const {
    return (this->counts > rhs.counts);
}


