//
// 
//

#ifndef META_DATACONTAINERS_H
#define META_DATACONTAINERS_H

#include "vector"
#include "stdexcept"
#include "set"
#include "iostream"

typedef std::set<std::pair<int,int>> gapDistribution;

template <typename T>
struct Count{
    Count(T key, int count);
    bool operator<(const Count<T>& rhs) const;
    bool operator>(const Count<T>& rhs) const;
    T key;
    int count;
};

template<typename T>
Count<T>::Count(T key, int count):key(key), count(count) {}

template<typename T>
bool Count<T>::operator<(const Count<T> &rhs) const {
    return count<rhs.count;
}

template<typename T>
bool Count<T>::operator>(const Count<T> &rhs) const{
    return count>rhs.count;
}


struct DataIndex{ // encodes a location in the sequence database
    DataIndex(int sequence_id, int event_id);
    int sequence_id;
    int event_id;
    bool operator<(const DataIndex& rhs) const;
    bool operator==(const DataIndex& rhs) const;
    bool operator>(const DataIndex& rhs) const;
    bool operator>=(const DataIndex& rhs) const;
    int operator-(const DataIndex& rhs) const;
    DataIndex& operator++();
};

struct PQEntryFindWin{ // priority queue entry for find win method
public:
    PQEntryFindWin(DataIndex start_index, int pattern_length);
    DataIndex current_index;
    int gaps = 0;
    int pattern_length;

    int getNextEventIndex();
    bool operator<(const PQEntryFindWin& rhs) const;
    void addPosition(DataIndex nextPosition);

    std::vector<DataIndex> getPositions();
private:

    int remaining_symbols;
    std::vector<DataIndex> positions;
};

struct PQEntryGapDistributionExtension{
public:
    PQEntryGapDistributionExtension(int gap_distribution_of, int relative_to,
                                    double distance);
    bool operator<(const PQEntryGapDistributionExtension& rhs) const;
    bool operator>(const PQEntryGapDistributionExtension& rhs) const;

    int gap_distribution_of;
//    int inserted_position;
    int relative_to;
    double distance;

    void print();

};



struct ReturnfindExistingAbsinAbs{
    ReturnfindExistingAbsinAbs(int direct_abstractions, int direct_children, int all_children){
        this->direct_abstractions = direct_abstractions;
        this->direct_children = direct_children;
        this->all_children = all_children;
    }
    int direct_abstractions;
    int direct_children;
    int all_children;
};

struct PQEntryRealCounts{
    PQEntryRealCounts(int pattern_e_index, int event_id, int counts);

    bool operator<(const PQEntryRealCounts rhs) const;
    bool operator>(const PQEntryRealCounts rhs) const;
    int pattern_e_index;
    int event_id;
    int counts;
};



#endif //META_DATACONTAINERS_H
