//
// 
//



#ifndef META_MODEL_H
#define META_MODEL_H


#include "list"
#include "stdexcept"
#include "map"
#include "queue"
#include "algorithm"
#include "iostream"
#include "set"
#include "float.h"

#include "Pattern.h"
#include "SequDatabase.h"
#include "utils.h"
#include "DataContainerSecond.h"
#include "AvailableWindowSet.h"

typedef Candidate candidate;
typedef std::list<candidate> candidateSet;
typedef std::map<int,Pattern*> patternSet;
typedef std::priority_queue<PQEntryGapDistributionExtension, std::vector<PQEntryGapDistributionExtension>, std::greater<PQEntryGapDistributionExtension>> pq_gap_dis_ext;
typedef std::priority_queue<PQEntryPatternOverlap, std::vector<PQEntryPatternOverlap>, std::greater<PQEntryPatternOverlap>> pq_pattern_overlap;

class Model {
public:
    Model(SequDatabase& db);
    ~Model();


    double getEncodedLength() const;

    WindowSet findWindows(Pattern& pattern, int budget);
    WindowSet findWindows(Pattern& pattern, int budget, AbstractionExtension current_abs_ext ,std::vector<AbstractionExtension> prev_abs_extensions);
    WindowSet greedyCover(WindowSet& candidate_windows);
    WindowSet greedyCover();

    int addAbstraction(Abstraction a, bool assignId = true);
    void extendAbstraction(int id, std::vector<int> new_events);
    void removeElementFromAbstraction(int abs_id, int element);
    void removeAbstraction(int id);
    std::map<int,Abstraction> getAbstractions();
    int getNewAbstractionId();
    void applyAbstractionExtensionList(std::vector<AbstractionExtension> extensions);
    void applyAbstractionExtension(AbstractionExtension extension);
    void undoAbstractionExtensionList(std::vector<AbstractionExtension> extension);
    void undoAbstractionExtension(AbstractionExtension extension);

    WindowSet& getCover();
    patternSet getSelection();

    void printSelection(bool ignore_singleton=false);
    std::string getPatternOutputString(bool ignore_singletons, bool incl_trans_p = false);
    std::string getAbstractionOutputString();


    void loadAbstractionsFromFile(const std::string& filePath);
    void loadPatternsFromFile(const std::string& filePath);
    void loadOtherPatternsFromFile(const std::string& filePath);
    void computePatternStructureForAll();


    std::vector<std::map<int,GapDistribution>> extendInterleaving(Pattern& pattern);
    Candidate extendByFollowing(Pattern& pattern);
    Candidate extendByProceeding(Pattern& pattern);
    ExtendCandidateReturn extendCandidate(Candidate& candidate);

    void search(bool in_search_pruning, bool prune_with_recover, bool print_steps);

    void printAbstractions();

    const std::map<int,int>& getUsageCounts();

    double ct_length;
    double cover_length;
    double fix_length;


private:


    SequDatabase db;
    patternSet selection;
    patternSet open;
    std::map<int,Abstraction> abstractions;
    double encoded_length;
    double fixed_cost;
    int patternCounter; //last added pattern id -  simple counter to keep track of patterns - not all values between 0 and max patternId have to be used
    int abstractionCounter;
    std::map<int,int> pattern_singleton_usage; //first ALPHABET SIZE elements singletons, remaining usage of patterns
    std::map<int,int> gap_counts;
    int sum_of_usages;
    WindowSet cover;
    AvailableWindowSet availableWindowSet;
    std::list<double> specificationProbabilities;

    std::map<int,int> usage_copy_diagnositics;

    void computeLength(bool ignore_spec_cost = false);
    double computeCoverLength(bool ignore_spec_cost = false);
    double computeCodeTableLength();
    void computeFixedCost();
    double computePatternLength(const Pattern& pattern);
    void recomputeUsages();

    ExtendCandidateReturn refinePatternInterleaving(Pattern& p);
    std::vector<ExtendCandidateReturn> refinePattern(Pattern& p);
    std::vector<ExtendCandidateReturn> refineCandidate(Candidate& pattern_candidate);
    std::pair<Candidate,Candidate> splitSingletonAndPatterns(Candidate& candidate);
    std::vector<ExtendCandidateReturn> patternCandidatesToPattern(Candidate candidate);
    void extendAbstractions();

    void addPatternToSelection(Pattern *p, bool assign_id);
    void removePatternFromSelection(Pattern* p);
    void removePatternFromSelection(int patternId);
    std::vector<Window*> conflictResolve(std::vector<Window*> conflictSet);

    ReturnfindExistingAbsinAbs findExistingAbsinAbs(int id, std::vector<AbstractionExtension>& prev_added, std::vector<AbstractionExtension>::iterator& limit);
    ReturnGainEstAbsExt estimateGainFromAbsExt(std::vector<AbstractionExtension> prev_abs_ext, AbstractionExtension& abstractionExtension, std::vector<int> zero_cost_abs_ext = {});
    double estimatePatternGain(PatternExtension pattern, extension_list counts);
    double estimatePatternGain(Pattern pattern, int pattern_frequ, UniqueValueCounter<int> frequ_updates);
    double estimatePatternGain(PatternExtension& pattern, extension_list  counts, std::vector<AbstractionExtension> prev_abs_ext ,AbstractionExtension& abstractionExtension);
    double estimatePatternGain(PatternExtension &pattern, std::vector<std::map<int, GapDistribution>> &counts_after_i,
                               std::vector<AbstractionExtension> prev_abs_ext,
                               AbstractionExtension &abstractionExtension,
                               std::vector<int> zero_cost_abs);
    double computeGain(Pattern &pattern, std::vector<AbstractionExtension> prev_abs_ext, bool ignore_spec_cost = false);
    double computeGain(Pattern &pattern, std::vector<AbstractionExtension> prev_abs_ext, Pattern &source_pattern);
    double computeGainAbsExt(std::vector<AbstractionExtension> abs_ext_list, Pattern &pattern);




    Abstraction& getAbstraction(int i);
    std::vector<int> getAllAbstractionsOfEvent(int event);
    std::vector<int> getAllAbstractionsOfEvent(int event, AbstractionExtension ext, std::vector<AbstractionExtension> extensions);
    std::vector<int> getAllChildrenOfEvent(int event); //if event abstraction returns all children else just given event
    std::vector<int>
    getAllChildrenOfEvent(int event, AbstractionExtension latest_ext, std::vector<AbstractionExtension> extensions,
                          std::set<int> seen, bool ignore_self = false);
    std::vector<int> getAllChildrenOfAbstraction(Abstraction& abstraction);
    std::vector<int> getAllChildrenOfAbsExt(AbstractionExtension ext, std::vector<AbstractionExtension> abst_ext_list);
    int otherAbstractions(Abstraction& abstraction);
    std::vector<int> findMatchingAbstraction(std::set<int> children);

    AbstractionExtension extendPattern(PatternExtension &p, int position, int extension, bool extend_abstraction, std::vector<AbstractionExtension> abstraction_extension_list);
    ExtendCandidateReturn extendPattern(Pattern& pattern, std::vector<std::map<int,GapDistribution>>& counts_after_i);

    ExtendCandidateReturn cutOutAbstractions(ExtendCandidateReturn candidate);


    double estimateGapFrequGain(Pattern& p, GapDistribution& distribution);

    void addAllGapDistributionToPQ(pq_gap_dis_ext &pq, Candidate &candidate,
                                   int for_element ,std::set<int> already_added);

    bool subSet(std::vector<int>& sub, std::vector<int>& par);
    int findSmallestSuperSetAbstraction(AbstractionExtension ext, std::vector<AbstractionExtension> abstractionExtensionList);

    void extendCandidateWithAbstractions(Candidate& candidate);

    int positionToRelativ(int pattern_length, int position);
    void update_pq_insert_positions(std::map<int,int>& pq_insert_positions, int new_position);
    double estimate_delta_updated_frequ(int new_pattern_frequ, UniqueValueCounter<int> frequ_updates);
    double estimate_specification_delta(std::vector<int>& old_f, std::vector<int>& new_f);
    double estimate_spec_delta_for_patterns(UniqueValueCounter<int> combined_pattern_frequ_updates);
    double ct_gain_by_zero_frequ(UniqueValueCounter<int> combined_pattern_frequ_updates);

    ExtendCandidateReturn estimateGainOnRealCounts(ExtendCandidateReturn addition_candidate, double gain_to_beat);
    void computePattenStructure(Pattern* pattern);
    std::pair<std::vector<int>, std::vector<std::pair<int,int>>> mergePatternsAndFlattenAbs();
    std::vector<int> prunePatterns();
    std::vector<int> pruneWithReCover();
    void removeUnusedAbstractions();
    bool abstractionUsed(int id);
    std::vector<std::pair<int,int>> replaceSymbolInPatterns(int to_replace,int with);
    void replaceBackPatterns(std::vector<std::pair<int,int>> changes, int with);
    void replaceSymbolInAbstractions(int to_replace,int with);
//    void replaceBackAbstractions()
    void replaceSymbolInOpen(int to_replace, int with);
    std::vector<std::pair<int,int>> replaceSymbol(patternSet&, int to_replace, int with);
    bool inChildren(int abs, int event);

    void reCoverData();
    int checkCover();
};

bool l_greater_r(Window* lhs, Window* rhs);

#endif //META_MODEL_H
