//
// 
//


#ifndef META_PATTERN_H
#define META_PATTERN_H

#include "vector"
#include "list"
#include "iostream"
#include "map"

#include "DataContainers.h"
#include "Window.h"
#include "utils.h"


enum LinkType {NONE, EQUALS, DEPENDENT_ON, count};

// key: (event_id 1, event_id 2) Contingency table between events --  value: counts of joined occurrences of events
typedef std::map<std::pair<int,int>,UniqueValueCounter<std::pair<int,int>>> contingency_table;

class Pattern { //is specific to one model, as it can use abstractions specified within the model.
public:
    Pattern(const std::vector<int>& events, std::vector<int> dependent_on, std::vector<LinkType> link_type);
    Pattern(int id, std::vector<int> events, std::vector<int> dependent_on, std::vector<LinkType> link_type);
    std::list<Window*>& getLocations();
    std::list<Window>& getCopyLocations();
    int getPatternUsageCounts();
    void addLocation(Window* w);
    void clearLocations();
    void copyLocations();

    void addWindow(Window w);
    void clearWindows_and_Contingency_tables();
    void allWindowsSet();
    void voidWindowSet();
    int allWindowsSize();
    int getMarginCounts(int i, int event);

    void replace(int position, int event, int dependent_on, LinkType linkType, bool splice = false);
    void splice( int position, int event,int dependent_on, LinkType linkType);
    void extendWithPattern(Pattern& pattern, int position);


    std::vector<int> getEvents() const;
    void replaceEvent(int new_event, int index);
    void removeEvent(int index);
    std::vector<int> getDependsOn();
    std::vector<LinkType> getLinkTypes();
    int getId();
    void setId(int id);

    int dependentOn(int i) const;
    LinkType kindOfDependence(int i) const;

    void setDependentOn(int i, int dependent_on);

    void clearUsageCounts();
    void incUsageCounts(int i,int conditionedOn, int event);
    int getUsageCount(int i, int conditionedOn, int event);
    int getSumUsageCount(int i, int conditionedOn);
    UniqueValueCounter<int> getUnconditionalUsageCounts(int i);

    contingency_table getContingencyTableFromLocationCopy();


    bool operator==(const Pattern& rhs) const;
    bool operator!=(const Pattern& rhs) const;
    bool operator>(const Pattern& rhs) const;
    void print();
    std::string getOutputString();
    std::string getTransPasString();

private:
    int id = -1;
    std::vector<int> events;
    std::vector<int> depends_on;
    std::vector<LinkType> link_type;

    std::list<Window*> locations; // windows used in cover
    std::list<Window> copy_location;

    bool allWindowsUpToDate = false; //idk about that, not even a guaranteed to be correct in case abs get extended windows change without modification to pattern.
    std::list<Window> allWindows; // all windows where pattern matches data

    std::vector<UniqueValueCounter<int>> margin_usage_counts;

    // key: (event_id 1, event_id 2) Contingency table between events --  value: counts of joined occurrences of events
    contingency_table contingency_tables_over_all_windows;

    std::vector<std::map<std::pair<int,int>,int>> usageCounts;



};

typedef std::map<int,GapDistribution> extension_list;

class Candidate{
public:
    /**
     *
     * @param pattern Pattern to be extended
     * @param position position of pattern to be extended
     * 0 appended on the front patter
     * length of pattern appended to the back
     * @param extensionList vector of possible extensions
     */

    Candidate(Pattern& pattern, int position, extension_list extensionList);

    extension_list getExtensions();
    Pattern* getCopyOfPattern();
    int getPosition();
    Pattern& getPattern();
    void addExtensionToList(int id, GapDistribution gd);

    void printExtensionList();
private:
    Pattern& pattern;
    int position;
    extension_list extensionList;
};

class Abstraction{
public:
    Abstraction(std::vector<int> children);
    Abstraction(int id, std::vector<int> children);
    std::vector<int> getChildren();
    int getId();
    void setId(int id);
    void extendAbstraction(int extension);
    void removeChildrenFromAbstraction(std::vector<int> removeSet);
    void killTheDoppelgaengers();
    void replaceSymbol(int to_replace, int with);
    std::string getOutputString();
    bool operator<(const Abstraction& rhs) const;
private:
    int id;
    std::vector<int> children;
};


#endif //META_PATTERN_H
