#include <iostream>
#include <getopt.h>
#include "limits"
#include <sys/stat.h>
#include <fstream>

#include "Model.h"
#include "SequDatabase.h"

#include "Pattern.h"
#include "GlobalData.h"

int main(int argc, char ** argv) {

    srand(1);

    std::cout.precision(std::numeric_limits<double>::max_digits10);

    static struct option longopts[] = {
            {"input",  required_argument, NULL, 1},
            {"output", required_argument, NULL, 2},
            {"windowsize", required_argument, NULL, 3},
            {"method", required_argument, NULL, 4},
            {"max-ext-factor", required_argument, NULL, 10},
            {"abstractions", optional_argument, NULL, 5},
            {"patterns", optional_argument, NULL, 6},
            {"in-search-pruning", optional_argument, NULL, 7},
            {"prune-re-cover", optional_argument, NULL, 8},
            {"disable-output", optional_argument, NULL, 9}

    };

    std::string input_file_path;
    std::string output_folder_path;
    std::string abstraction_file_path;
    std::string pattern_file_path;
    std::string method;
    int windowsize;
    int window_ext_factor;
    bool in_search_pruning = false;
    bool prune_re_cover = false;
    bool output = true;

    int parameter;
    int long_index =0;
    while ((parameter = getopt_long_only(argc,argv,"", longopts,&long_index)) != -1){
        switch (parameter) {
            case 1:
                input_file_path = optarg;
                break;
            case 2:
                output_folder_path = optarg;
                break;
            case 3:
                windowsize = std::atoi(optarg);
                break;
            case 4:
                method = optarg;
                break;
            case 5:
                abstraction_file_path= optarg;
                break;
            case 6:
                pattern_file_path = optarg;
                break;
            case 7:
                in_search_pruning = true;
                break;
            case 8:
                prune_re_cover = true;
                break;
            case 9:
                output = false;
                break;
            case 10:
                window_ext_factor = std::atoi(optarg);
                break;
            default:
                std::cout << "Default Case" << std::endl;
        }
    }

    GlobalData::window_budget = windowsize;
    GlobalData::max_pattern_extension_factor = window_ext_factor;


    SequDatabase db = SequDatabase(input_file_path);
    GlobalData::db = &db;
    Model m = Model(db);

    double null_model_cost = m.getEncodedLength();
    if (output){
        std::cout << "Singleton Cover Length: ";
        std::cout << null_model_cost << std::endl;
        m.printSelection();
    }


    if (mkdir(output_folder_path.c_str(),0777) == -1){
        std::cerr << "No output folder created, might already exist." << std::endl;
    }

    if (method == "search"){
        m.search(in_search_pruning, prune_re_cover, output);

    }else if (method == "model-given"){
        if (!abstraction_file_path.empty()){
            m.loadAbstractionsFromFile(abstraction_file_path);
        }
        if (!pattern_file_path.empty()){
            m.loadPatternsFromFile(pattern_file_path);
            std::cout << "Given Pattern Cover Length: ";
            std::cout << m.getEncodedLength() << "=" << m.ct_length << " + " << m.cover_length << " + " << m.fix_length << std::endl;

        }
    }
    else if(method == "model-given-serach-dep"){
        if (!abstraction_file_path.empty()){
            m.loadAbstractionsFromFile(abstraction_file_path);
        }
        if (!pattern_file_path.empty()){
            m.loadPatternsFromFile(pattern_file_path);
            m.computePatternStructureForAll();
            std::cout << "Given Pattern Cover Length: ";
            std::cout << m.getEncodedLength() << "=" << m.ct_length << " + " << m.cover_length << " + " << m.fix_length << std::endl;

        }
    }
    else if (method == "merge-given"){
        if (!pattern_file_path.empty()){
            m.loadOtherPatternsFromFile(pattern_file_path);
        }
    }
    else if (method == "search-given"){
        if (!pattern_file_path.empty()){
            m.loadOtherPatternsFromFile(pattern_file_path);
        }
        m.search(in_search_pruning, prune_re_cover, output);
    }
    else{
        throw std::runtime_error("Invalid method");
    }

    if (method != "model-given") {
        std::ofstream pattern_file(output_folder_path + "/patterns.txt");
        pattern_file << m.getPatternOutputString(true) << std::endl;
        pattern_file.close();

        std::ofstream abstraction_file(output_folder_path + "/abstractions.txt");
        abstraction_file << m.getAbstractionOutputString() << std::endl;
        abstraction_file.close();
    }

//    std::ofstream pattern_trans_file(output_folder_path + "/patterns_inc_transition_counts.txt");
//    pattern_trans_file << m.getPatternOutputString(true, true) << std::endl;
//    pattern_trans_file.close();

//   // int combined_usage = getUsageCounts().find(p1.first)->second + getUsageCounts().find(p2.first)->second;

    std::ofstream bits_file(output_folder_path + "/bits.txt");
    bits_file << "bits under null model, bits under model" << std::endl;
    bits_file << null_model_cost << std::endl;
    bits_file << m.getEncodedLength() << std::endl;
    bits_file.close();

    std::cout << "Results saved in " << output_folder_path << std::endl;


    if (output){
        m.printSelection();
        std::cout << "Cover Length: ";
        std::cout << m.getEncodedLength() << std::endl;
    }


    return 0;
}
