//
// 
//

#ifndef META_UTILS_H
#define META_UTILS_H

#include <iostream>
#include "math.h"
#include "stdexcept"
#include "vector"
#include "map"
#include "algorithm"

#include "DataContainers.h"
#include "wasserstein.h"


double universal_integer_encoding(int i);
inline double logFact(const uint32_t n);
double logBinom(const uint32_t n, const uint32_t k);
bool checkOverlap(std::vector<DataIndex>& p1, std::vector<DataIndex>& p2);

template <typename T>
class UniqueValueCounter{
public:
    void incCount(T key, int value);
    int getCount(T key);
    std::vector<Count<T>> getSortedCounts();

    std::map<T,int> getCounts();
    std::vector<T> getKeys();
    void print();
private:
    std::map<T,int> counts;
};

template<typename T>
void UniqueValueCounter<T>::incCount(T key, int value) {
    auto countsIt = counts.template find(key);
    if (countsIt == counts.end()){
        counts[key] = value;
    }else{
        counts[key] += value;
    }
}

template<typename T>
int UniqueValueCounter<T>::getCount(T key) {
    auto countsIt = counts.find(key);
    if (countsIt != counts.end()){
        return counts[key];
    }else{
        return 0;
    }
}

template<typename T>
std::vector<Count<T>> UniqueValueCounter<T>::getSortedCounts() {
    std::vector<Count<T>> sorted_counts;
    for (auto c:counts){
        sorted_counts.push_back(Count<T>(c.first,c.second));
    }
    std::sort(sorted_counts.begin(),sorted_counts.end(),std::greater<Count<T>>());
    return sorted_counts;
}

template<typename T>
std::map<T, int> UniqueValueCounter<T>::getCounts() {
    return counts;
}

template<typename T>
void UniqueValueCounter<T>::print() {
    std::cout << "Counts:" << std::endl;
    for (auto e:counts){
        std::cout << e.first << " - " << e.second << std::endl;
    }
    std::cout << std::endl;
}

template<typename T>
std::vector<T> UniqueValueCounter<T>::getKeys() {
    std::vector<T> keys;
    for (auto e:counts){
        keys.push_back(e.first);
    }
    return keys;
}


// works on absolut counts
class GapDistribution{
public:
    void incDeltaFrequency(int delta, int value);
    int getDeltaFrequency(int delta);
    int getTotalCounts();
    double getDeltaProbability(int delta);
    void removeDeltaFromDistribution(int delta);
    std::vector<int> getAllDeltas();
    double divergence(GapDistribution& other);
    double divergence_frqu_adjsted(GapDistribution& other);
    double getMStat();
    void print();

private:
    int total_counts = 0;
    std::map<int,int> distribution;
    double m_stat;
    bool m_stat_correct;
    double kullback_leibler_discrete(GapDistribution& other);
    double jensen_shannon_discrete(GapDistribution& other);
    double earth_move_distance(GapDistribution& other);
    double getMean();
    double getMode();
    double getMedian();
};

#endif //META_UTILS_H
